from abc import ABC, abstractmethod

from .log_parser import LogEntry
import re
import logging

class AbstractIpBlockExecution(ABC):
    @abstractmethod
    def begin_execute(self):
        pass

    @abstractmethod
    def block(self,log: LogEntry):
        pass

    @abstractmethod
    def end_execute(self):
        pass


class FileBasedUWFBlock(AbstractIpBlockExecution):
    """
        write generated firewall rules into a file
    """

    def __init__(self, destinate_path: str):
        self.__destinate_path = destinate_path
        self.__blocked_item = []

    def begin_execute(self):
        self.__blocked_item = []
        pass

    def block(self,log: LogEntry):
        for n in re.split(",\\s*",log.network):
            ufw_block_cmd = f"ufw deny from {n.strip()} to any"
            logging.debug("use command %s", ufw_block_cmd)
            self.__blocked_item.append(f"{ufw_block_cmd}\n")
        pass

    def end_execute(self):
        with open(self.__destinate_path, 'w') as f:
            f.write("#!/bin/bash\n")
            f.writelines(self.__blocked_item)
        logging.info("Block %d IPs", len(self.__blocked_item))
        pass


import sys, os
import argparse
import logging
import configparser
import glob

from typing import List, Dict
from pprint import pprint, pformat

from . config_parser import ParserConfigException, \
    VERBOSITY, LOG_LEVELS, CONF_FILE, \
    LOG_FILES, LOG_PATTERN, DATABASE_PATH, \
    JUDGMENT, RULES, \
    BOT_REQUEST, MAX_REQUEST, INTERVAL_SECONDS, \
    EXECUTION, SCRIPT, \
    parse_config_file

from . import log_parser
from . import judgment
from . import execution

# work-flow
# 1. suche alle IP in Logfile nach Merkmale eines Angriff
# 2. generiert UFW Befehlen aus der IPs im Schritte 1
# 3. gebe die Befehlen aus, oder sonstiges weiter Verarbeitung

# CLI options:

# [judgment]

# []

_parser = None


def main():
    global _parser
    argv = sys.argv
    parser = argparse.ArgumentParser()
    parser.add_argument(f"{CONF_FILE}", type=str, nargs="?",
                        help="Configuration file, configuration must be given by a configuration file "
                        f"(the positional argument {CONF_FILE}). The Debug level can be changed by "
                        f"optional argument --{VERBOSITY}.")

    parser.add_argument("-v", f"--{VERBOSITY}", default="INFO",
                        choices=LOG_LEVELS,
                        help="how much information is printed out during processing log files")
    _parser = parser
    cli_arg = vars( parser.parse_args(argv[1:]) )
    verbosity = cli_arg[VERBOSITY]
    apply_log_config(verbosity)
    file_based_config = parse_config_file(cli_arg[CONF_FILE])
    if VERBOSITY in cli_arg: file_based_config[VERBOSITY] = verbosity
    if logging.getLogger("root").isEnabledFor(logging.DEBUG): logging.debug(pformat(file_based_config))
    validate_config(file_based_config)
    try:
        analyse_log_files(file_based_config)
    except judgment.JudgmentException as ex:
        print(ex, file=sys.stderr)


def apply_log_config(verbosity: str):
    log_level = logging.getLevelName(verbosity)
    logging.getLogger().setLevel(level=log_level)
    logging.info("Verbosity: %s %d", verbosity, log_level)


def validate_config(config: Dict):
    if LOG_FILES not in config:
        raise ParserConfigException("Log files are not configured")


#TODO: test this method
def analyse_log_files(config: Dict):
    log_files = expand_log_files(config[LOG_FILES])
    judge = construct_judgment(config)
    executor = execution.FileBasedUWFBlock(config[EXECUTION][0][RULES][SCRIPT])
    executor.begin_execute()
    log_pattern = config[LOG_PATTERN]
    for file_path in log_files:
        logging.info("Analyse file %s", file_path)
        logs = log_parser.parse_log_file(file_path, log_pattern)
        for log in logs:
            logging.debug("Process `%s'", log)
            if judgment.is_ready_blocked(log, config[DATABASE_PATH]):
                logging.info("IP %s is ready blocked", log.ip_str)
            elif judge.should_deny(log):
                network = judgment.lookup_ip(log.ip_str)
                log.network = network
                executor.block(log)
    executor.end_execute()


def expand_log_files(config_log_file: List[str]) -> List:
    log_files = []
    for p in config_log_file:
        expand_path = glob.glob(p)
        logging.debug("expand glob '%s' to %s", p, expand_path)
        if len(expand_path) == 0:
            logging.warn("Glob path '%s' cannot be expanded to any real path", p)
        log_files = log_files + expand_path
    return log_files


def construct_judgment(config) -> judgment.AbstractIpJudgment:
    judgments_chain = config[JUDGMENT] if JUDGMENT in config else []
    if len(judgments_chain) < 1:
        _parser.error(f"At least one entry in {JUDGMENT} must be configured")
    list_of_judgments = []
    for judge in judgments_chain:
        list_of_judgments += [judgment_by_name(judge, config)]
    logging.info("Use %d judgments", len(list_of_judgments))
    return judgment.ChainedIpJudgment( config[DATABASE_PATH], list_of_judgments)


def judgment_by_name(judge, config):
    name = judge['name']
    rules = judge[RULES]
    if name == "path-based-judgment":
        bot_request_path = rules[BOT_REQUEST] if BOT_REQUEST in rules else []
        if len(bot_request_path) > 0:
            return judgment.PathBasedIpJudgment(bot_request_path)
        else:
            _parser.error(f"At least one path in {BOT_REQUEST} must be configured if Judgment {name} is used")
    elif name == "time-based-judgment":
        if DATABASE_PATH in config:
            database_path = config[DATABASE_PATH]
            max_request = judge[MAX_REQUEST] if MAX_REQUEST in judge else 500
            interval = judge[INTERVAL_SECONDS] if INTERVAL_SECONDS in judge else 60
            return judgment.TimeBasedIpJudgment(database_path, max_request, interval)
        else:
            _parser.error(f"A SQLite database ({DATABASE_PATH}) must be configured in global section if Judgment {name} is used")
    else:
        raise ParserConfigException(f"Unknown judgment {name}")


#############################################################################
#############################################################################
#############################################################################
def init_db():
    global _parser
    argv = sys.argv
    parser = argparse.ArgumentParser()
    parser.add_argument("db_path",
                        help="Path to an Sqlite Database")
    _parser = parser
    cli = parser.parse_args(argv[1:])
    db_path = cli.db_path
    judgment.init_database(db_path)
    pass


import gzip
import ipaddress
from datetime import datetime
import logging
from typing import List

import magic

DATETIME_FORMAT_PATTERN = '%Y-%m-%d %H:%M:%S.%f%z'


class LogEntry:
    """
    represents a Log Entry with following attribute:
        * 'ip': the remote IP
        * 'time': timestamp of this log entry
        * 'status': HTTP-response status of this log
        * 'request': the first line of the HTTP-request or None if not available
        * 'byte': response length in Byte
        * 'user': remote-user or None if not available
    """

    def __init__(self, log_file: str,
                 line: int,
                 ip: int = 0,
                 network: str = None,
                 time: datetime = None,
                 status: int = 0,
                 request: str = None,
                 byte: int = 0,
                 user: str = None
                 ):
        self.__ip = ip
        self.__network = network
        self.__time = time
        self.__status = status
        self.__request = request
        self.__byte = byte
        self.__user = user
        self.__log_file = log_file
        self.__line = line

    @property
    def ip(self) -> int:
        """
            the IP of the log entry, represents as an interger
        :return: the ip
        """
        return self.__ip

    @ip.setter
    def ip(self, ip: str or int):
        """
            set the ip of the entry, the argument can be a string, or an integer. If the argument
            is a string, it will be converted into am integer
            # TODO: validate IP
        :param ip:
        :return:
        """
        if type(ip) == str:
            self.__ip = LogEntry.ip_to_int(ip)
        elif type(ip) == int:
            self.__ip = ip
        pass

    @property
    def network(self) -> str:
        """
            the network of ip of this Log, may be None
        :return:
        """
        return self.__network

    @network.setter
    def network(self, network: str):
        """
            set network for the ip of this Log
        :param network:
        :return:
        """
        self.__network = network

    @property
    def time(self) -> datetime:
        """
            time of this log entry
        :return:
        """
        return self.__time

    @time.setter
    def time(self, time: str or datetime):
        """
            set time for this log entry. The argument can be a String or an instance of `datetime'
        :param time: time of this log entry
        :return:
        """
        self.__time = time

    @property
    def status(self) -> int:
        return self.__status

    @status.setter
    def status(self, status: int or str):
        self.__status = int(status)
        pass

    @property
    def request(self) -> str:
        return self.__request

    @property
    def byte(self) -> int:
        return self.__byte

    @byte.setter
    def byte(self, byte: str or int):
        self.__byte = int(byte)

    @request.setter
    def request(self, request: str):
        self.__request = request

    @property
    def user(self) -> str:
        return self.__user

    @user.setter
    def user(self, user: str):
        self.__user = user

    @property
    def iso_time(self) -> str:
        return self.__time.strftime(DATETIME_FORMAT_PATTERN)

    @property
    def ip_str(self):
        return int_to_ip(self.__ip)

    @property
    def log_file(self) -> str:
        return self.__log_file

    @log_file.setter
    def log_file(self, log_file: str):
        self.__log_file = log_file

    @property
    def line(self) -> int:
        return self.__line

    @line.setter
    def line(self, line:int):
        self.__line = line

    def __getitem__(self, item):
        if item == "ip":
            return self.__ip
        elif item == "time":
            return self.__time
        elif item == "status":
            return self.__status
        elif item == "request":
            return self.__request
        elif item == "user":
            return self.__user
        elif item == "byte":
            return self.__byte
        elif item == "log_file":
            return self.__log_file
        elif item == "line":
            return self.__line
        else:
            raise KeyError("{} does not have property {}".format(self.__class__.__name__, item))

    def __setitem__(self, key, value):
        if key == "ip":
            self.ip = value
        elif key == "time":
            self.time = value
        elif key == "status":
            self.status = value
        elif key == "request":
            self.request = value
        elif key == "user":
            self.user = value
        elif key == "byte":
            self.byte = value
        elif key == "log_file":
            self.log_file = value
        elif key == "line":
            self.line = value
        else:
            raise KeyError("{} does not have property {}".format(self.__class__.__name__, key))

    def __str__(self):
        return "{}:{}: {} {} {} {} {} {}".format(
            self.log_file,
            self.line,
            int_to_ip(self.ip),
            self.iso_time,
            self.user,
            self.request,
            self.status,
            self.byte
        )


def parse_log_file(log_file_path, log_pattern) -> List[LogEntry]:
    """
    gibt eine Liste von IP in `tomcat_access_path'-Datei zurück, welche als Angriff
    eingestuft wird.

    @param log_file_path: path zur Access Logfile von Tomcat, in der Regel ist die Datei ${CATALINA_BASE}/logs/localhost_access${date}.txt

    @param log_pattern: Wie in der Doku von Tomcat beschrieben (https://tomcat.apache.org/tomcat-9.0-doc/config/valve.html#Access_Logging),
    z.B:

    %h %{X-Forwarded-For}i %l %u %t &quot;%r&quot; %s %b

    """

    log_pattern = log_pattern.replace('&quot;', '"').split(' ')
    logs = []
    num_of_line = 0
    file_reader_fn = open_log_file_fn(log_file_path)
    with file_reader_fn(log_file_path) as logfile:
        line = logfile.readline()
        num_of_line += 1
        while line:
            log_entry = parser_tomcat_log_line(log_file_path, num_of_line, line, log_pattern)
            logs.append(log_entry)
            line = logfile.readline()
    logging.debug("parsed %d lines", num_of_line)
    return logs


def open_log_file_fn(file_path):
    file_type = magic.from_file(file_path)
    if file_type.startswith('gzip compressed data'):
        return lambda fp: gzip.open(fp, 'rt', encoding="utf-8",errors='ignore')
    else:
        return lambda fp: open(fp)


def parser_tomcat_log_line(log_file_path: str, num_of_line: int, log_line: str, pattern: str) -> LogEntry:
    entry = LogEntry(log_file_path, num_of_line)
    line_idx = 0
    host = None
    proxy_host = None

    def _parser_word(log_line, start_idx):
        word = ''
        i = start_idx
        for i in range(start_idx, len(log_line)):
            c = log_line[i]
            if c != ' ':
                word += c
            else:
                i += 1
                break

        return (word, i)

    def _parser_sentence(log_line, start_idx, begin_quote='"', end_quote='"'):
        sentence = ''
        if log_line[start_idx] != begin_quote:
            raise TypeError("Expected string")
        i = start_idx
        for i in range(start_idx + 1, len(log_line)):
            c = log_line[i]
            if c == end_quote:
                i += 2  # '"' and ' '
                break
            else:
                sentence += c
        return (sentence, i)

    for idx, value in enumerate(pattern):
        if value == '%h':
            (entry_value, line_idx) = _parser_word(log_line, line_idx)
            host = entry_value
            proxy_host = None
            logging.debug("Capture host %s", host)
        elif value == '%{X-Forwarded-For}i':
            (entry_value, line_idx) = _parser_word(log_line, line_idx)
            proxy_host = entry_value
            logging.debug("Capture proxy_host %s", proxy_host)
            host = None
        elif value == '%u':
            (entry_value, line_idx) = _parser_word(log_line, line_idx)
            entry.user = entry_value
        elif value == '%t':
            (entry_value, line_idx) = _parser_sentence(log_line, line_idx, begin_quote='[', end_quote=']')
            entry.time = datetime.strptime(entry_value, '%d/%b/%Y:%H:%M:%S %z')
        elif value == '"%r"':
            (entry_value, line_idx) = _parser_sentence(log_line, line_idx)
            entry.request = entry_value
        elif value == '%s':
            (entry_value, line_idx) = _parser_word(log_line, line_idx)
            entry.status = int(entry_value)
        elif value == '%b':
            (entry_value, line_idx) = _parser_word(log_line, line_idx)
            entry.byte = int(entry_value)
        else:
            (entry_value, line_idx) = _parser_word(log_line, line_idx)
            logging.debug("ignore pattern %s", value)

    # update ip at the last attribute
    logging.debug("host=%s proxy=%s", host, proxy_host)
    try:
        if proxy_host is not None:
            logging.debug("use proxy_host %s", proxy_host)
            entry.ip = ip_to_int(proxy_host)
        else:
            logging.debug("use host %s", host)
            entry.ip = ip_to_int(host)

        return entry
    except ipaddress.AddressValueError as ex:
        raise LogParserException(log_file_path, num_of_line, str(ex), errors=ex)


def ip_to_int(ip: str) -> int:
    """
    (2^(8*3))*a + (2^(8*2))*b + (2^8)*c + d
    :param ip:
    :return:
    """
    return int(ipaddress.IPv4Address(ip))


def int_to_ip(ip: int) -> str:
    return str(ipaddress.IPv4Address(ip))


class LogParserException(Exception):
    def __init__(self,log_file, line, message, errors=None):
        self.message = "({},{}) {}".format(log_file, line, message)
        self.errors = errors
        super(LogParserException, self).__init__(message)